#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import copy
import datetime
import logging
import math
import os
import re
from pathlib import Path

from pycapo import CapoConfig

from . import NGASUtil as ng
from . import almautils, archiveutils, projectutils, proposalutils
from .evlasciencedatamodel import SDMBinaryTable, SDMXMLTable
from .RealFastMetadata import RealFastMetadata
from .schema import create_session
from .schema.model import (
    AlmaOus,
    AncillaryProduct,
    Author,
    Configuration,
    DataDescription,
    ExecutionBlock,
    File,
    Filegroup,
    Intent,
    Polarization,
    Project,
    RealfastExecutionBlock,
    Scan,
    ScienceProduct,
    Subscan,
)

LOG = logging.getLogger("ingestion")

session = create_session("SDM")

config = CapoConfig()
site_name = config.getstring("archive-ingestion.NGASSite")
cluster_name = config.getstring("archive-ingestion.NGASCluster")


def persist(sdm, nocal, additional_metadata, collection_metadata, science_products, ingestion_path, ngas_cluster):
    """
    Persist an sdm (Science Data Model) to the metadata db, including the NGAS index to be
    used by the Data Fetcher via SOLR
    :param sdm: Science Data Model for an EVLA fileset
    :param nocal:
    :param additional_metadata: Addition metadata
    :param collection_metadata: Collection metadata
    :param science_products: Science products for output group
    :param ingestion_path Path of files to be ingested
    :param ngas_cluster: ngas cluster
    :return: the project
    """

    sdmtables = sdm.sdmtables
    bdffiles = sdm.bdffiles
    fileset = sdm.fileset

    # The set of XML tables contained in the SDM
    maintable = sdm.sdmtables["Main"].table.getroot()
    sbsummarytable = sdm.sdmtables["SBSummary"].table.getroot()
    execblocktable = sdm.sdmtables["ExecBlock"].table.getroot()
    scantable = sdm.sdmtables["Scan"].table.getroot()
    subscantable = sdm.sdmtables["Subscan"].table.getroot()
    polarizationtable = sdm.sdmtables["Polarization"].table.getroot()
    spectralwindowtable = sdm.sdmtables["SpectralWindow"].table.getroot()
    datadescriptiontable = sdm.sdmtables["DataDescription"].table.getroot()
    configtable = sdm.sdmtables["ConfigDescription"].table.getroot()
    fieldtable = sdm.sdmtables["Field"].table.getroot()
    sourcetable = sdm.sdmtables["Source"].table.getroot()
    receivertable = sdm.sdmtables["Receiver"].table.getroot()

    try:
        window_receivers = {
            row.spectralWindowId.text: row.frequencyBand.text.split("_")[-1] for row in receivertable.row
        }
    except:
        LOG.warning("Unable to obtain spectral window -> receiver mapping")
        window_receivers = {}

    # Persist the NGAS metadata
    project, parentfilegroup, bdffilemap, sdm_ingest_files = persistngasmetadata(
        sdmtables, bdffiles, fileset, sbsummarytable, execblocktable, subscantable
    )
    session.commit()

    # Persist the science metadata
    output_products, ingest_files = persistsciencemetadata(
        fileset,
        bdffilemap,
        project,
        parentfilegroup,
        sbsummarytable,
        execblocktable,
        scantable,
        subscantable,
        maintable,
        configtable,
        datadescriptiontable,
        polarizationtable,
        spectralwindowtable,
        fieldtable,
        sourcetable,
        nocal,
        window_receivers,
        additional_metadata,
        collection_metadata,
        science_products,
        ingestion_path,
        ngas_cluster,
    )
    session.commit()
    return project, output_products, sdm_ingest_files + ingest_files


def persistngasmetadata(sdmtables, bdffiles, fileset, sbsummarytable, execblocktable, subscantable):
    """
    Persist the NGAS index entries
    :param sdmtables: the set of sdm tables
    :param bdffiles: the set of bdf files
    :param fileset: the file set being persisted
    :param sbsummarytable: SBSummary table
    :param execblocktable: ExecBlock table
    :param subscantable: SubScan table
    :return: project, filegroup, map of bdffiles
    """

    # Persist the project
    ingest_files = []
    project = persistproject(sbsummarytable, execblocktable, subscantable)

    if project.authors is None or len(project.authors) == 0:
        # Persist the authors if not an existing project
        authors = persistauthors(project)
        for author in authors:
            session.add(author)
        project.authors = authors

    # Persist the filegroup
    parentfilegroup, filegroup = createfilegroup(project)

    # Persist files for sdm tables
    filegroupsize = 0
    for sdmtable in sdmtables:
        file = createsdmtable(sdmtables, filegroup, sdmtable, execblocktable.row.telescopeName)
        session.add(file)
        filegroupsize += file.filesize
        ingest_files.append((file, file.filesize))

    # Persist files for bdf files
    bdffilemap = {}
    for bdffile in bdffiles:
        if execblocktable.row.telescopeName == "EVLA" and "X" in bdffile.entityid:
            LOG.warning(f"Observation {fileset} has invalid BDF {bdffile.entityid}")
            continue
        file = createbdffile(bdffile, filegroup, execblocktable.row.telescopeName)
        session.add(file)
        filegroupsize += file.filesize
        bdffilemap[bdffile.entityid] = file
        ingest_files.append((file, file.filesize))

    filegroup.datasize = filegroupsize
    session.add(filegroup)
    parentfilegroup.datasize = -1
    session.add(parentfilegroup)
    session.add(project)
    return project, parentfilegroup, bdffilemap, ingest_files


def persistsciencemetadata(
    fileset,
    bdffilemap,
    project,
    filegroup,
    sbsummarytable,
    execblocktable,
    scantable,
    subscantable,
    maintable,
    configtable,
    datadescriptiontable,
    polarizationtable,
    spectralwindowtable,
    fieldtable,
    sourcetable,
    nocal,
    window_receivers,
    additional_metadata,
    collection_metadata,
    output_science_products,
    ingestion_path,
    ngas_cluster,
):
    """
    Persist the science meta data from the SDM tables
    :param bdffilemap:
    :param project:
    :param filegroup: Top level filegroup for the observation
    :param sbsummarytable:
    :param execblocktable:
    :param scantable:
    :param subscantable:
    :param maintable:
    :param configtable:
    :param datadescriptiontable:
    :param polarizationtable:
    :param spectralwindowtable:
    :param fieldtable:
    :param sourcetable:
    :param nocal:
    :param window_receivers:
    :param additional_metadata: additional metadata
    :param collection_metadata: collection metadata
    :param output_science_products: Output group science products
    :param ingestion_path: Path of files to ingest
    :param ngas_cluster: ngas cluster
    :return:
    """
    # Persist ExecBlock
    execblockdict = {}
    science_products = []
    ingest_files = []
    for execblock in execblocktable.row:
        execblockobject, science_product, files_to_ingest = persistexecblock(
            fileset,
            project,
            filegroup,
            sbsummarytable,
            execblocktable,
            subscantable,
            nocal,
            additional_metadata,
            collection_metadata,
            output_science_products,
            ingestion_path,
            ngas_cluster,
        )
        ingest_files = ingest_files + files_to_ingest
        execblockdict[execblock.execBlockId.text] = execblockobject
        session.add(science_product, execblockobject)
        science_products.append(science_product)
    session.commit()

    for execblock in execblocktable.row:
        execblockobject = execblockdict[execblock.execBlockId.text]
        scans = getscans(execblock.execBlockId, scantable)
        # Persist scans
        bandset = set()
        for scan in scans:
            try:
                scanobject = persistscan(
                    scan,
                    execblockobject,
                    filegroup,
                    maintable,
                    fieldtable,
                    sourcetable,
                    spectralwindowtable,
                    configtable,
                    datadescriptiontable,
                    polarizationtable,
                )
                session.add(scanobject)

                # Persist subscans
                subscans = getsubscans(scanobject.ost_scan_id, subscantable)
                for subscan in subscans:
                    subscanobject, bands = persistsubscan(
                        bdffilemap,
                        execblockobject,
                        scanobject,
                        subscan,
                        scan.scanIntent.text,
                        maintable,
                        configtable,
                        datadescriptiontable,
                        polarizationtable,
                        spectralwindowtable,
                        fieldtable,
                        sourcetable,
                        execblocktable,
                        window_receivers,
                    )
                    bandset = bandset.union(bands)
                    session.add(subscanobject)
            except ValueError:
                continue
        execblockbands = " ".join(sorted(bandset))
        execblockobject.band_code = execblockbands
        session.add(science_product, execblockobject)
    return science_products, ingest_files


def lookup_external_project(telescope, sbsummarytable):
    title = ""
    abstract = ""
    legacyid = ""
    schedulingblocktype = sbsummarytable.row.sbType.text
    authors = []

    # Assign the proprietary duration
    if schedulingblocktype == "OBSERVER":
        proprietaryduration = 365
    else:
        proprietaryduration = 0

    if telescope == "ALMA":
        projectid = sbsummarytable.row.projectUID.EntityRef.get("entityId")
        almaproject = almautils.getproject(projectid)
        authors.extend(almautils.getauthors(projectid))

        title = almaproject.title
        abstract = almautils.getabstract(projectid)
        legacyid = None
        projectcode = almaproject.code
        optprojectid = None
    else:
        # EVLA stuff
        projectid = getid(sbsummarytable.row.projectUID.EntityRef)

        if "X" in projectid:
            # Project id was not available in the sdm table, set to 'Operations'
            projectcode = "Operations"
            optprojectid = 0
        else:
            optproject = projectutils.getproject(projectid)
            if optproject is None:  # No match in the OPT
                projectcode = "Not in OPT   "
                optprojectid = 0
            else:
                projectcode = optproject.projectcode
                optprojectid = optproject.id

                # All VLASS project codes should be mapped onto the VLASS proposal
                if optproject.projectcode.startswith("VLASS"):
                    proposal = proposalutils.getproposal("VLASS")
                else:
                    proposal = proposalutils.getproposal(optproject.projectcode)

                authors.extend(projectutils.getauthors(projectid))

                if proposal is not None:
                    title = proposal.TITLE
                    abstract = proposal.ABSTRACT
                    legacyid = proposal.LEGACY_ID
                    if proposal.PROPOSAL_TYPE == "Director's Discretionary Time":
                        proprietaryduration = round(365 / 2, 0)

    return ExternalProject(
        projectcode, title, abstract, proprietaryduration, legacyid=legacyid, optprojectid=optprojectid, authors=authors
    )


def persistproject(sbsummarytable, execblocktable, subscantable):
    """
    Persist project
    :param sbsummarytable:
    :param execblocktable:
    :param subscantable:
    :return:
    """
    # If the project is already persisted, just update the last_addition date
    # and the start time and end time
    telescope = execblocktable.row.telescopeName
    extproject = lookup_external_project(telescope, sbsummarytable)
    existingproject = archiveutils.getproject(extproject.project_code)
    if existingproject is not None:
        # Get the min, max for the exec block
        minstarttime, maxendtime = getminmaxtime(execblocktable.row.execBlockId, subscantable)
        # Propagate to project if appropriate
        if minstarttime < existingproject.starttime:
            existingproject.starttime = minstarttime
        if maxendtime > existingproject.endtime:
            existingproject.endtime = maxendtime
        existingproject.last_addition = datetime.datetime.utcnow()
        return existingproject

    totaltime = totalobservingtime(execblocktable)
    # If semester proposal = 365
    # If test = 0
    # If semester and DDT = 365 / 2
    minstarttime, maxendtime = getminmaxtime(execblocktable.row.execBlockId, subscantable)
    # last_addition - update when another run for this project
    lastaddition = datetime.datetime.utcnow()
    project = Project(
        project_code=extproject.project_code,
        opt_project_id=extproject.opt_project_id,
        legacy_id=extproject.legacy_id,
        title=extproject.title,
        abstract=extproject.abstract,
        total_observation_time=totaltime,
        last_addition=lastaddition,
        starttime=minstarttime,
        endtime=maxendtime,
        proprietary_duration=extproject.proprietary_duration,
    )
    if telescope == "ALMA":
        project.authors = extproject.authors

    return project


class ExternalProject:
    def __init__(self, projectcode, title, abstract, proprietaryduration, **kwargs):
        self.project_code = projectcode
        self.opt_project_id = kwargs["optprojectid"]
        self.legacy_id = kwargs["legacyid"]
        self.title = title
        self.abstract = abstract
        self.proprietary_duration = proprietaryduration
        self.authors = kwargs["authors"] if "authors" in kwargs else []


def persistauthors(project):
    """
    Persist authors
    :param projectid:
    :return:
    """
    saveauthors = []
    authors = projectutils.getauthors(project.opt_project_id)
    for author in authors:
        existingauthor = archiveutils.getauthor(project.project_code, author.globalid)
        if existingauthor is not None:
            saveauthors.append(existingauthor)
        else:
            # Now get the user information from the PST using the globalid
            pstuser = proposalutils.getpstuser(author.globalid)
            saveauthors.append(
                Author(
                    project_code=project.project_code,
                    username=pstuser.personAuthentication.personName,
                    firstname=pstuser.firstName,
                    lastname=pstuser.lastName,
                    pst_person_id=author.globalid,
                    is_pi=projectutils.ispi(project.project_code, author.globalid),
                )
            )
    return saveauthors


def createfilegroup(project):
    """
    Create filegroup
    :param project:
    :return:
    """
    parentfilegroup = Filegroup(project=project, type="execution_block")
    filegroup = Filegroup(parent_filegroup=parentfilegroup, type="rawdata")
    return parentfilegroup, filegroup


def createsdmtable(sdmtables, filegroup, sdmtable, telescope):
    """
    Create file for sdm table
    :param sdmtables:
    :param filegroup:
    :param sdmtable:
    :param telescope:
    :return:
    """
    file_path = filegroup.type
    if isinstance(sdmtables[sdmtable], SDMXMLTable):
        ngas_id = sdmtables[sdmtable].entityid.replace("://", "___").replace("/", "_") + ".sdm"
        filename = sdmtables[sdmtable].tablename + ".xml"
    elif isinstance(sdmtables[sdmtable], SDMBinaryTable):
        ngas_id = sdmtables[sdmtable].entityid.replace("://", "___").replace("/", "_") + ".bin"
        filename = sdmtables[sdmtable].tablename + ".bin"
    else:
        raise Exception("Invalid SDM table type")
    filesize = sdmtables[sdmtable].size / 1000
    format = "sdm"
    type = "raw"

    file = File(
        file_path=file_path,
        ngas_id=ngas_id,
        ngas_location=site_name,
        ngas_cluster=cluster_name,
        filegroup1=filegroup,
        filename=filename,
        filesize=filesize,
        format=format,
        type=type,
        ingestion_time=datetime.datetime.utcnow(),
    )

    if telescope == "ALMA":
        file.ngas_location = "NAASC"
        file.ngas_cluster = "NAASC"

    return file


TELESCOPE_SIZE_CORRECTION_FACTORS = {"EVLA": 1000, "ALMA": 1000}


def createbdffile(bdffile, filegroup, telescope):
    """
    Create file for bdf
    :param bdffile:
    :param filegroup:
    :return:
    """
    file_path = filegroup.type
    if telescope == "EVLA" and "X" in bdffile.entityid:  # dummy uid; I think this is deprecated
        ngas_id = None
        filename = bdffile.entityid  # The id given in the XML
        filesize = 0
    elif ":///" in bdffile.entityid:  # EVLA
        filename = bdffile.entityid.replace(":///", "____").replace("/", "_")
        ngas_id = filename + ".bdf"
        # the filesize sometimes appears to be -99, what's that about?
        filesize = int(bdffile.filesize) / TELESCOPE_SIZE_CORRECTION_FACTORS.get(telescope, 1)
        if filesize < 0:
            LOG.warning(f"Negative filesize {filesize} for filename {filename}")
    elif "://" in bdffile.entityid:  # ALMA
        filename = bdffile.entityid.replace("://", "___").replace("/", "_")
        ngas_id = filename + ".bdf"
        # the filesize sometimes appears to be -99, what's that about?
        filesize = int(bdffile.filesize) / TELESCOPE_SIZE_CORRECTION_FACTORS.get(telescope, 1)
        if filesize < 0:
            LOG.warning(f"Negative filesize {filesize} for filename {filename}")
    format = "bdf"
    type = "raw"
    file = File(
        file_path=file_path,
        ngas_id=ngas_id,
        ngas_location=site_name,
        ngas_cluster=cluster_name,
        filegroup1=filegroup,
        filename=filename,
        filesize=filesize,
        format=format,
        type=type,
        ingestion_time=datetime.datetime.utcnow(),
    )
    return file


def getid(entityref):
    entityid = entityref.get("entityId")
    idloc = entityid.rfind("/")
    return entityid[idloc + 1 :]


def totalobservingtime(execblock):
    # TODO units?
    observingtime = 0.0
    for row in execblock.row:
        starttime = row.startTime / 86400e09
        stoptime = row.endTime / 86400e09
        observingtime += stoptime - starttime
    return observingtime


def persistexecblock(
    fileset,
    project,
    filegroup,
    sbsummarytable,
    execblocktable,
    subscantable,
    nocal,
    additional_metadata,
    collection_metadata,
    output_science_products,
    ingestion_path,
    ngas_cluster,
):

    files_to_ingest = []
    ost_exec_block_id = getid(execblocktable.row.execBlockUID.EntityRef)
    if "X" in ost_exec_block_id:
        ost_exec_block_id = 0
    calibration_level = "0"
    telescope = execblocktable.row.telescopeName.text
    configuration = None
    if len(execblocktable.row.findall("configName")) > 0:
        configuration = execblocktable.row.configName.text
    scheduling_block_id = getid(sbsummarytable.row.sbSummaryUID.EntityRef)
    if "X" in scheduling_block_id:
        scheduling_block_id = 0
    minstarttime, maxendtime = getminmaxtime(execblocktable.row.execBlockId, subscantable)
    schedulingblocktype = sbsummarytable.row.sbType.text
    if nocal:
        calibration_status = "Do Not Calibrate"
    else:
        calibration_status = None

    science_product_locator = ng.generate_science_product_locator(telescope.lower(), "execblock")

    if collection_metadata is not None and collection_metadata.isRealFast():
        realfast_metadata = RealFastMetadata(collection_metadata)
        execblock = RealfastExecutionBlock(
            ost_exec_block_id=ost_exec_block_id,
            filegroup=filegroup,
            calibration_level=calibration_level,
            telescope=telescope,
            configuration=configuration,
            scheduling_block_id=scheduling_block_id,
            project=project,
            ngas_fileset_id=fileset,
            starttime=minstarttime,
            endtime=maxendtime,
            calibration_status=calibration_status,
            scheduling_block_type=schedulingblocktype,
            science_product_locator=science_product_locator,
            transient_ra=realfast_metadata.get_transient_ra(),
            transient_ra_error=realfast_metadata.get_transient_ra_error(),
            transient_dec=realfast_metadata.get_transient_dec(),
            transient_dec_error=realfast_metadata.get_transient_dec_error(),
            transient_snr=realfast_metadata.get_transient_snr(),
            transient_dm=realfast_metadata.get_transient_dm(),
            transient_dm_error=realfast_metadata.get_transient_dm_error(),
            preaverage_time=realfast_metadata.get_preaverage_time(),
            rfpipe_version=realfast_metadata.get_rfpipe_version(),
            prefs_id=realfast_metadata.get_prefs_id(),
            rf_qa_label=realfast_metadata.get_rf_qa_label(),
            rf_qa_zero_fraction=realfast_metadata.get_rf_qa_zero_fraction(),
            rf_qa_visibility_noise=realfast_metadata.get_rf_qa_visibility_noise(),
            rf_qa_image_noise=realfast_metadata.get_rf_qa_image_noise(),
            portal_candidate_ids=realfast_metadata.get_portal_candidate_ids(),
        )
    else:
        execblock = ExecutionBlock(
            ost_exec_block_id=ost_exec_block_id,
            filegroup=filegroup,
            calibration_level=calibration_level,
            telescope=telescope,
            configuration=configuration,
            scheduling_block_id=scheduling_block_id,
            project=project,
            ngas_fileset_id=fileset,
            starttime=minstarttime,
            endtime=maxendtime,
            calibration_status=calibration_status,
            scheduling_block_type=schedulingblocktype,
            science_product_locator=science_product_locator,
        )
    science_product = ScienceProduct(
        science_product_locator=science_product_locator,
        science_product_type="Execution Block",
        metadata_ingestion_date=datetime.datetime.utcnow(),
        metadata_ingestion_version=1,
        external_system="VLA Operations",
        external_name=fileset,
        filegroup=filegroup,
    )

    execblock.project.science_products.append(science_product)

    if "ancillary_products" in output_science_products[0].keys():
        ancillary_products = output_science_products[0]["ancillary_products"]
        for anc_product in ancillary_products:
            LOG.debug(f"Ancillary product: {anc_product}")
            product_file = Path(ingestion_path) / anc_product["filename"]
            ancillary_product_locator = ng.generate_science_product_locator(telescope.lower(), "ancillary_product")
            ancillary_product_filegroup = Filegroup(
                type="ancillary_product",
                datasize=-1,
            )
            if anc_product["type"] == "thumbnail_image":
                thumbnail_path = str(Path(ingestion_path) / product_file.name)
                preview_path = str(Path(calculate_thumbnail_path(thumbnail_path)[0]).parent)
            else:
                preview_path = None
            ancillary_product_file = File(
                ngas_id=ng.generate_uuids_for_ngas(telescope, "ancillary_product", product_file.suffix[1:]),
                ngas_location=site_name,
                ngas_cluster=ngas_cluster,
                filename=product_file.name,
                filesize=product_file.stat().st_size,
                format=product_file.suffix[1:],
                type=anc_product["type"],
                checksum=ng.sha256(product_file),
                checksum_type="SHA256",
                ingestion_time=datetime.datetime.utcnow(),
                filegroup1=ancillary_product_filegroup,
                preview_storage_path=preview_path,
            )

            files_to_ingest.append((ancillary_product_file, ancillary_product_file.filesize))

            ancillary_product = AncillaryProduct(
                ancillary_product_locator=ancillary_product_locator,
                ancillary_product_type=anc_product["type"],
                filegroup=ancillary_product_filegroup,
                science_product_locator=science_product_locator,
            )
            session.add(ancillary_product_filegroup)
            session.add(ancillary_product_file)
            session.add(ancillary_product)

    # For everything other than ALMA, just null out the alma_ous_uid
    if telescope == "ALMA":
        mous, gous, sous = almautils.getalmaoushierarchy(execblock.ngas_fileset_id)
        createalmaoushierarchy(project.project_code, mous, gous, sous)
        execblock.alma_ous_id = mous
        science_product.external_system = "ALMA Operations"
        science_product.external_name = fileset
    else:
        execblock.alma_ous_id = None
        science_product.external_system = "VLA Operations"
        science_product.external_name = fileset

    return execblock, science_product, files_to_ingest


def createalmaoushierarchy(project_code, mousid, gousid, sousid):
    """
    Create the OUS hierarchy for an ALMA execblock
    :param project_code:
    :param mousid:
    :param gousid:
    :param sousid:
    :return:
    """

    schedblock_name = almautils.getschedblockname(mousid)
    if archiveutils.ousexists(mousid, "MOUS"):
        # Nothing needed. Hierarchy already exists, just add eb
        LOG.info(f"MOUS exists, do nothing")
        return
    elif archiveutils.ousexists(gousid, "GOUS"):
        LOG.info(f"No MOUS, but GOUS exists, add MOUS")
        mous = AlmaOus(alma_ous_id=mousid, parent_ous_id=gousid, ous_type="MOUS", schedblock_name=schedblock_name)
        session.add(mous)
    elif archiveutils.ousexists(sousid, "SOUS"):
        LOG.info(f"MOUS and GOUS don't exist, but SOUS exists. Add MOUS and GOUS")
        gous = AlmaOus(alma_ous_id=gousid, parent_ous_id=sousid, ous_type="GOUS")
        session.add(gous)
        mous = AlmaOus(alma_ous_id=mousid, parent_ous_id=gousid, ous_type="MOUS", schedblock_name=schedblock_name)
        session.add(mous)
    else:
        LOG.info("SOUS doesn't exist. Create MOUS, GOUS, and SOUS")
        sous = AlmaOus(alma_ous_id=sousid, project_code=project_code, ous_type="SOUS")
        session.add(sous)
        gous = AlmaOus(alma_ous_id=gousid, parent_ous_id=sousid, ous_type="GOUS")
        session.add(gous)
        mous = AlmaOus(alma_ous_id=mousid, parent_ous_id=gousid, ous_type="MOUS", schedblock_name=schedblock_name)
        session.add(mous)


def getscans(execblockid, scantable):
    # Get only scans associated with this
    scanlist = []
    for scan in scantable.row:
        if scan.execBlockId == execblockid:
            scanlist.append(scan)

    return scanlist


def getsubscans(scanid, subscantable):
    subscanlist = []
    for subscan in subscantable.row:
        if subscan.scanNumber.text == scanid:
            subscanlist.append(subscan)
    return subscanlist


def persistscan(
    scan,
    execblock,
    filegroup,
    maintable,
    fieldtable,
    sourcetable,
    spectralwindowtable,
    configtable,
    datadescriptiontable,
    polarizationtable,
):
    ost_scan_id = scan.scanNumber.text
    minfrequency, maxfrequency, minbandwidth, maxbandwidth, polarizationcode = getaggregatedatadescriptionfields(
        ost_scan_id, maintable, configtable, datadescriptiontable, polarizationtable, spectralwindowtable
    )
    if "OBSERVE" in scan.scanIntent.text:
        obstype = "OBS"
    elif "CALIBRATE" in scan.scanIntent.text:
        obstype = "CAL"
    else:
        obstype = "OTHER"
    scanobject = Scan(
        ost_scan_id=ost_scan_id,
        execution_block=execblock,
        filegroup=filegroup,
        max_bandwidth=maxbandwidth,
        min_bandwidth=minbandwidth,
        polarization_code=polarizationcode,
        max_frequency=maxfrequency,
        min_frequency=minfrequency,
    )
    return scanobject


def getdatadescriptions(scanid, subscanid, maintable, configtable):
    maintablerow = getmaintablerow(maintable, scanid, subscanid)
    configdescriptionrow = getconfigtablerow(configtable, maintablerow.configDescriptionId)
    dataDescriptions = configdescriptionrow.dataDescriptionId.text.split()[2:]
    return dataDescriptions


def persistdatadescription(
    scanid, subscanid, subscanobject, datadescriptionid, datadescriptiontable, polarizationtable, spectralwindowtable
):
    frequency, bandwidth, polarizationcode, polarizations = getdatadescriptionfields(
        scanid, subscanid, datadescriptionid, datadescriptiontable, polarizationtable, spectralwindowtable
    )
    polarization = archiveutils.getpolarization(polarizationcode)
    # datadescription = archiveutils.getdatadescription(bandwidth, frequency, polarization.polarization_id, session)
    # if datadescription is None:
    datadescription = DataDescription(
        bandwidth=bandwidth, frequency=frequency, polarization_id=polarization.polarization_id
    )
    session.add(datadescription)
    return datadescription


def persistsubscan(
    bdffilemap,
    execblock,
    scan,
    subscan,
    scanIntent,
    maintable,
    configtable,
    datadescriptiontable,
    polarizationtable,
    spectralwindowtable,
    fieldtable,
    sourcetable,
    execblocktable,
    window_receivers,
):
    backend = "none"
    intent = " ".join(scanIntent.split(" ")[2:])
    starttimeMJD = datetoMJD(subscan.startTime)
    endtimeMJD = datetoMJD(subscan.endTime)
    subscanid = subscan.subscanNumber.text

    if intent == "CALIBRATE_POINTING":
        obstype = "POINT"
    elif intent == "UNSPECIFIED":
        obstype = "UNKWN"
    elif "OBSERVE_TARGET" in intent:
        obstype = "OBS"
    elif "CALIBRATE" in intent:
        obstype = "CAL"
    elif intent == "none":
        obstype = "TRACK"
    else:
        obstype = "OTHER"

    exposure_time = round((endtimeMJD - starttimeMJD) * 86400, 2)  # 4 decimal places
    if subscan.find("numIntegration") is not None and subscan.numIntegration != 0:
        integration_time = round(exposure_time / subscan.numIntegration, 3)
    elif subscan.find("numberIntegration") is not None and subscan.numberIntegration != 0:
        integration_time = round(exposure_time / subscan.numberIntegration, 3)
    else:
        integration_time = 0
    ra, dec = getraanddec(maintable, fieldtable, sourcetable, scan.ost_scan_id, subscan.subscanNumber)
    # TODO receiver id and build receiver table
    bands = getbands(
        scan.ost_scan_id,
        subscan.subscanNumber,
        maintable,
        configtable,
        datadescriptiontable,
        spectralwindowtable,
        execblocktable,
        window_receivers,
    )
    receiver_id = 0

    subscanobject = Subscan(
        scan=scan,
        file_id=10,
        ost_subscan_id=subscanid,
        obstype=obstype,
        starttime=starttimeMJD,
        endtime=endtimeMJD,
        sourcename="",
        sourcetype="STAR",
        ra=ra,
        dec=dec,
        exposure_time=exposure_time,
        integration_time=integration_time,
        receiver_id=receiver_id,
        backend=backend,
        intent=intent,
    )
    # Get list of data descriptions for this subscan / config
    datadescriptionlist = getdatadescriptions(scan.ost_scan_id, subscan.subscanNumber, maintable, configtable)
    # Build data_descripions rows
    maintablerow = getmaintablerow(maintable, scan.ost_scan_id, subscan.subscanNumber)
    configuration = re.sub("\w*_", "", maintablerow.configDescriptionId.text)
    configobject = archiveutils.getconfiguration(execblock.execution_block_id, int(configuration), session)
    if configobject is None:
        # Not found, create a new one
        configobject = Configuration(configuration=int(configuration), execution_block=execblock)
        for datadescription in datadescriptionlist:
            datadescriptionobject = persistdatadescription(
                scan.ost_scan_id,
                subscan.subscanNumber,
                subscanobject,
                datadescription,
                datadescriptiontable,
                polarizationtable,
                spectralwindowtable,
            )
            datadescriptionobject.configuration = configobject
            session.add(datadescriptionobject)
        session.add(configobject)

    # XML change
    bdffileid = None
    if maintablerow.find("dataUID") is not None:
        bdffileid = maintablerow.dataUID.EntityRef.get("entityId")
    elif maintablerow.find("dataOid") is not None:
        bdffileid = maintablerow.dataOid.EntityRef.get("entityId")
    else:
        LOG.critical("Invalid XML element in Main table")

    if execblocktable.row.telescopeName == "EVLA" and "X" in bdffileid:
        subscanobject.file = None
    else:
        subscanobject.file = bdffilemap[bdffileid]
    fieldtablerow = getfieldrow(fieldtable, maintablerow.fieldId)
    sourcetablerow = getsourcerow(sourcetable, fieldtablerow.sourceId)
    subscanobject.sourcename = sourcetablerow.sourceName.text
    subscanobject.configuration = configobject
    return subscanobject, bands


def getaggregatedatadescriptionfields(
    scanid, maintable, configtable, datadescriptiontable, polarizationtable, spectralwindowtable
):
    minfrequency, maxfrequency, minbandwidth, maxbandwidth, pcode = 0, 0, 0, 0, 240
    frequencies = []
    bandwidths = []
    polarizationset = set()
    subscans = []
    datadescriptions = []
    for row in maintable.row:
        if row.scanNumber.text == scanid:
            subscans.append(row.subscanNumber.text)
    for subscan in subscans:
        datadescriptions = getdatadescriptions(scanid, subscan, maintable, configtable)
    for datadescription in datadescriptions:
        pcode = 0
        frequency, bandwidth, polarizationcode, polarizations = getdatadescriptionfields(
            scanid, subscans, datadescription, datadescriptiontable, polarizationtable, spectralwindowtable
        )
        frequencies.append(frequency)
        bandwidths.append(bandwidth)
        polarizationset = polarizationset.union(polarizations)
    for polarizationelement in polarizationset:
        pcode += archiveutils.getpolarizationcode(polarizationelement)
    minfrequency = min(frequencies)
    maxfrequency = max(frequencies)
    minbandwidth = min(bandwidths)
    maxbandwidth = max(bandwidths)

    return minfrequency, maxfrequency, minbandwidth, maxbandwidth, pcode


def getdatadescriptionfields(
    scanid, subscanid, dataDescriptionId, datadescriptiontable, polarizationtable, spectralwindowtable
):
    datadescriptionrow = getdatadescriptionrow(datadescriptiontable, dataDescriptionId)
    spectralwindowrow = getspectralwindowrow(spectralwindowtable, datadescriptionrow.spectralWindowId)
    frequency = spectralwindowrow.refFreq.text
    bandwidth = spectralwindowrow.totBandwidth.text
    polarizationrow = getpolarizationrow(polarizationtable, datadescriptionrow.polOrHoloId)
    polarization = polarizationrow.corrType.text

    polarizations = ", ".join(polarization.split()[2:])
    polarizationlist = polarization.split()[2:]
    polarizationcode = archiveutils.getpolarizationcode(polarizations)

    return float(frequency), float(bandwidth), int(polarizationcode), polarizationlist


def getbands(
    scanid,
    subscanid,
    maintable,
    configtable,
    datadescriptiontable,
    spectralwindowtable,
    execblocktable,
    window_receivers,
):
    # TODO use this for bands

    bandsset = set()
    datadescriptionlist = getdatadescriptions(scanid, subscanid, maintable, configtable)
    # Build data_descripions rows
    for datadescription in datadescriptionlist:
        datadescriptionrow = getdatadescriptionrow(datadescriptiontable, datadescription)
        # NB. The ALMA method here is ostensibly the right way to do this.
        # However, the CBE in the past guessed what the receiver should be, and
        # wrote that guess into the SDM in the Receivers table. So, the value
        # in the Receivers table cannot be trusted in all cases for EVLA, hence,
        # the inference is made by the spectral window's name, which apparently
        # is correct in all cases for EVLA.
        if execblocktable.row.telescopeName == "ALMA":
            # make sure we have a receiver band
            if datadescriptionrow.spectralWindowId.text in window_receivers:
                bandsset.add(window_receivers[datadescriptionrow.spectralWindowId.text])
            else:
                LOG.error("Unable to locate any receivers for this spectral window!")
        else:
            spectralwindowrow = getspectralwindowrow(spectralwindowtable, datadescriptionrow.spectralWindowId)
            band = re.sub("\w*_", "", spectralwindowrow.name.text)
            band = re.sub("#\w*", "", band)
            bandsset.add(re.sub("#\w*", "", band))
    return bandsset


# Methods to return row matching id
def getmaintablerow(maintable, scanid, subscanid):
    for row in maintable.row:
        if int(row.scanNumber.text) == int(scanid) and int(row.subscanNumber.text) == int(subscanid):
            return row
    return None


def getconfigtablerow(configtable, configid):
    for row in configtable.row:
        if row.configDescriptionId == configid:
            return row
    return None


def getdatadescriptionrow(datadescriptiontable, datadescriptionid):
    for row in datadescriptiontable.row:
        if row.dataDescriptionId == datadescriptionid:
            return row
    return None


def getpolarizationrow(polarizationtable, polarizationid):
    for row in polarizationtable.row:
        if row.polarizationId == polarizationid:
            return row
    return None


def getspectralwindowrow(spectralwindowtable, spectralwindowid):
    for row in spectralwindowtable.row:
        if row.spectralWindowId == spectralwindowid:
            return row
    return None


def getfieldrow(fieldtable, fieldId):
    for row in fieldtable.row:
        if row.fieldId == fieldId:
            return row
    return None


def getsourcerow(sourcetable, sourceId):
    for row in sourcetable.row:
        if row.sourceId == sourceId:
            return row
    return None


def getraanddec(maintable, fieldtable, sourcetable, scanid, subscanid):
    mainrow = getmaintablerow(maintable, scanid, subscanid)
    fieldrow = getfieldrow(fieldtable, mainrow.fieldId)
    sourcerow = getsourcerow(sourcetable, fieldrow.sourceId)
    radeclist = sourcerow.direction.text.split(" ")
    return radeclist[2], radeclist[3]


def datetoMJD(subscantime):
    iat_utc = 0.0
    asdmMjd = 54500.0

    asdmMjd = (subscantime + iat_utc) / 86400.0e9
    return asdmMjd


def getminmaxtime(execblockid, subscantable):
    minstarttime, maxendtime = math.inf, -math.inf

    for row in subscantable.row:
        if row.execBlockId == execblockid:
            if row.startTime < minstarttime:
                minstarttime = row.startTime
            if row.endTime > maxendtime:
                maxendtime = row.endTime
    return datetoMJD(minstarttime), datetoMJD(maxendtime)


def filesdump(ingest_files):
    for file in ingest_files:
        LOG.info(f"File name: {file[0].filename} Size: {file[1]}")


def dump_bdf_files(bdffiles):
    for bdffile in bdffiles:
        LOG.info(f"BDF file name: {bdffile.entityid} Size: {bdffile.filesize}")
