#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Various 'n' sundry utilities for our tests """

# pylint: disable=E0401, R1721

from pathlib import Path
from typing import List

import pytest
from ingest_envoy.manifest_components import (
    WEBLOG_FILENAME,
    AncillaryProduct,
    OutputScienceProduct,
)
from ingest_envoy.utilities import AncillaryProductType

# --------------------------------
#    EVLA CAL manifest test data
# --------------------------------
EVLA_CAL_INPUT_FILENAMES = ["20A-346_2021_07_23_T13_37_08.376.tar", WEBLOG_FILENAME]
UNWANTED = ["ignore_me.fits", "just_a_lotta_nothing", "uninteresting_metadata.xml"]


@pytest.fixture(scope="function")
def ingest_path(tmpdir: Path) -> Path:
    """
    Make an "ingestion path" for tests

    :param tmpdir: temporary home for ingestion location
    :return:
    """

    # cast is necessary because otherwise we get a LocalPath, which doesn't work
    fake_ingest_path = Path(tmpdir / "ingestion")
    fake_ingest_path.mkdir()
    return fake_ingest_path


def find_example_manifest(manifest_name: str) -> Path:
    """
    Get this example manifest for comparison with one we've generated in a test.

    :param manifest_name: unique file identifier
    :return: full path to the manifest file
    """
    filename = manifest_name + ".json"
    for file in Path.cwd().rglob(filename):
        return file

    raise FileNotFoundError(filename)


def populate_fake_evla_cal_ingest_path(staging_dir: Path) -> List[Path]:
    """
    Create a directory containing fake calibration products, plus other stuff
    that we -don't- want to ingest.

    :param staging_dir: our temporary dir
    :return:
    """

    files = []
    filenames = [filename for filename in EVLA_CAL_INPUT_FILENAMES]
    for filename in UNWANTED:
        filenames.append(filename)

    for filename in filenames:
        path = staging_dir / filename
        path.touch()
        files.append(path)

    return files


# -----------------------------
#    Image manifest test data
# -----------------------------

EXAMPLE_MANIFEST_FILE = find_example_manifest("image_manifest_tmpx_ratuqh")

ADDITIONAL_METADATA_FILENAME = "aux_image_metadata.json"

PRIMARY_BEAM_ANCILLARY = AncillaryProduct(
    type=AncillaryProductType.PB_FITS, filename="oussid.1-93305_sci.L_band.cont.I.pb.tt0.fits"
)
CLEAN_MASK_ANCILLARY = AncillaryProduct(
    type=AncillaryProductType.FITS_MASK, filename="oussid.1-93305_sci.L_band.cont.I.mask.fits"
)
OUTPUT_GROUP_SCIENCE_PRODUCT_ANCILLARIES = [PRIMARY_BEAM_ANCILLARY, CLEAN_MASK_ANCILLARY]

OUTPUT_SCIENCE_PRODUCT = OutputScienceProduct(
    type=AncillaryProductType.FITS,
    filename="oussid.1-93305_sci.L_band.cont.I.tt0.fits",
    ancillary_products=OUTPUT_GROUP_SCIENCE_PRODUCT_ANCILLARIES,
)

# input files
WEBLOG_ANCILLARY = AncillaryProduct(type=AncillaryProductType.PIPELINE_WEBLOG_TYPE, filename=WEBLOG_FILENAME)
PIPELINE_AF_ANCILLARY = AncillaryProduct(
    type=AncillaryProductType.PIPELINE_ARTIFACTS,
    filename="pipeline_artifacts_2021_08_04T15_46_02.tar",
)
# an output file
INGESTION_AF_ANCILLARY = AncillaryProduct(
    type=AncillaryProductType.INGESTION_ARTIFACTS,
    filename="ingestion_artifacts_2021_08_04T01_57_08.564.tar",
)
ANCILLARY_PRODUCTS = [WEBLOG_ANCILLARY, INGESTION_AF_ANCILLARY, PIPELINE_AF_ANCILLARY]

STAGING_DIR_FILES = [
    "aux_image_metadata.json",
    "oussid.1-93305_sci.L_band.cont.I.tt0.fits",
]


def populate_fake_tmpx_ratuqh_ingest_path(staging_source_dir: Path) -> List[Path]:
    """
    make a bunch of fake files that should result in the example manifest

    :return:
    """
    fake_files_to_create = [ADDITIONAL_METADATA_FILENAME]

    for product in OUTPUT_GROUP_SCIENCE_PRODUCT_ANCILLARIES:
        fake_files_to_create.append(product.filename)

    fake_files_to_create.append(PIPELINE_AF_ANCILLARY.filename)
    fake_files_to_create.append(WEBLOG_ANCILLARY.filename)

    fake_files_to_create.append(OUTPUT_SCIENCE_PRODUCT.filename)

    files = []
    for filename in fake_files_to_create:
        file = staging_source_dir / filename
        file.touch()
        files.append(file)

    assert len(files) == len(fake_files_to_create)

    return files
