#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" This is the workflow service. """

import json
import logging
import os
import pathlib
import stat
import subprocess
from pathlib import Path
from tempfile import mkdtemp
from typing import Dict, List, Union

# pylint: disable=E0401, W1203

import requests
import transaction
from messaging.messenger import MessageSender
from messaging.router import Router, on_message
from pycapo import CapoConfig
from requests import Response

from workspaces.system.services.remote_processing_service import CapoInjector
from workspaces.workflow.enum import ArchiveWorkflows, WorkflowRequestState
from workspaces.workflow.message_architect import (
    ArchiveMessageArchitect,
    WorkflowMessageArchitect,
)
from workspaces.workflow.schema import Workflow, WorkflowRequest, WorkflowRequestFile
from workspaces.workflow.services.interfaces import WorkflowInfoIF, WorkflowServiceIF

logger = logging.getLogger(__name__)


class WorkflowServiceRESTClient(WorkflowServiceIF):
    """This is the workflow service REST client."""

    def __init__(self):
        self.url = CapoConfig().settings("edu.nrao.workspaces.WorkflowSettings").serviceUrl

    def execute(self, request: WorkflowRequest):
        """
        Execute a workflow request.

        :param request: the workflow request to be executed
        :return: the response as JSON
        """
        response = requests.post(
            f"{self.url}/workflows/{request.workflow_name}/requests/{request.workflow_request_id}/submit"
        )
        logger.info(
            "Got result %s with type %s and body %s",
            response,
            type(response),
            response.content,
        )
        return response.json()

    def attach_file_to_request(self, request: WorkflowRequest, filename: str, content: bytes) -> Response:
        """
        Add a file to this workflow request.

        :param request: the workflow request
        :param filename: name of file to be added
        :param content: file contents
        :return: response to the PUT()
        """
        response = requests.put(
            f"{self.url}/workflows/{request.workflow_name}"
            f"/requests/{request.workflow_request_id}"
            f"/files/{filename}",
            data=content,
        )
        return response

    def retrieve_file_content(self, name: str, request_id: int, filename: str):
        """
        Retrieve the conent of a workflow file
        :param name: the workflow name
        :param request_id: id of parent workflow request
        :param filename: file to find
        :return: dict containing file content
        """

        response = requests.get(f"{self.url}/workflows/{name}/requests/{request_id}/files/{filename}")
        return response.content.decode()

    def create_workflow_request(self, workflow: Union[str, Workflow], argument: Dict) -> WorkflowRequest:
        """
        Create a workflow request using the supplied arguments.

        :param workflow: workflow to be requested
        :param argument: arguments to the workflow request
        :return: the WorkflowRequest
        """

        # 1. Handle the type ambiguity with the workflow argument
        workflow_name = workflow if isinstance(workflow, str) else workflow.workflow_name

        # 2. Make the request
        result = requests.post(
            f"{self.url}/workflows/{workflow_name}/requests/create",
            json=argument,
        ).json()

        return WorkflowRequest.from_json(result)

    def announce_qa(self, workflow_request_id: int, msg_type: str):
        """
        Announce QA event

        :param workflow_request_id: id of workflow
        :param msg_type: pass/fail/ready
        :return:
        """
        requests.post(f"{self.url}/workflows/requests/{workflow_request_id}/qa/{msg_type}")

    def ingest(self, request: WorkflowRequest):
        """
        Ingest results for previously run workflow into NGAS and archive
        :param request: completed workflow request to ingest
        :return:
        """
        requests.post(f"{self.url}/workflows/{request.workflow_name}/requests/{request.workflow_request_id}/ingest")


class WorkflowService(WorkflowServiceIF):
    """
    Executes workflows; should be a freestanding service.
    """

    # The next few things that need to happen here are in response to
    # WorkflowEvents that we receive from wf_monitor as the workflow
    # execution evolves. So we have to set up listening at some point
    # in this class
    def __init__(self, info: WorkflowInfoIF, message_handler: "WorkflowMessageHandler"):
        self.info = info
        self.message_router = message_handler.message_router
        self.messenger = message_handler.messenger
        self.archive_messenger = message_handler.archive_messenger
        self.processing_settings = CapoConfig().settings("edu.nrao.workspaces.ProcessingSettings")

    def attach_file_to_request(self, request: WorkflowRequest, filename: str, content: bytes):
        """
        Add a file to a request; complain if the filename is bad.

        :param request: the request to which the file will be added
        :param filename: name of the file
        :param content: file contents

        :return:
        """
        forbidden = self._get_forbidden_templates_list(request.workflow_name)
        if filename not in forbidden:
            if ArchiveWorkflows.is_archive_wf(request.workflow_name) and filename == "metadata.json":
                content_dict = json.loads(content.decode())
                content_dict["workflowName"] = request.workflow_name
                content_dict["data_location"] = request.argument["data_location"]
                content = json.dumps(content_dict, indent=4).encode()

            self.info.save_file(request=request, filename=filename, content=content)
        else:
            logger.info(f"{filename} is a protected file name.")

    def retrieve_file_content(self, request_id: int, filename: str):
        wf_req = self.info.lookup_workflow_request(request_id)
        for file in wf_req.files:
            if file.filename == filename:
                return file.content.decode()

    def create_workflow_request(self, workflow, argument) -> WorkflowRequest:
        """
        Create a workflow request from the supplied arguments.

        :param workflow: the desired workflow
        :param argument: arguments to workflow request
        :return: the request that's been created
        """
        return self.info.create_workflow_request(workflow, argument)

    def send_carta_url_to_aat(self, workflow_request_id: int, carta_url: str) -> Dict[str, any]:
        """
        Send AMQP message with CARTA url included to the AAT AMQP exchange

        :param workflow_request_id: ID of carta workflow that wants to send this message
        :param carta_url: JSON blob with CARTA URL
        """
        logger.info(f"SENDING CARTA MESSAGE to AAT Request Handler for request #{workflow_request_id}!")
        wf_request = self.info.lookup_workflow_request(workflow_request_id)
        routing_key = f"ws-workflow.carta-instance-ready.{workflow_request_id}"
        carta_url_msg = ArchiveMessageArchitect(
            routing_key=routing_key, request=wf_request, carta_url=carta_url
        ).compose_message("carta_ready")
        self.archive_messenger.send_message(**carta_url_msg)
        return carta_url_msg

    def send_carta_url(self, workflow_request_id: int, carta_url: str):
        """
        Send AMQP message with CARTA URL to Workspaces.

        :param workflow_request_id: ID of CARTA workflow that wants to send this message
        :param carta_url: JSON blob containing CARTA URL
        :return:
        """
        logger.info(f"SENDING CARTA MESSAGE to Workspaces for request #{workflow_request_id}!")
        wf_request = self.info.lookup_workflow_request(workflow_request_id)
        carta_url_msg = WorkflowMessageArchitect(request=wf_request, carta_url=carta_url).compose_message("carta_ready")
        self.messenger.send_message(**carta_url_msg)

    def send_do_not_calibrate(self, request_id: int, sdm_id: str):
        """
        Tell the archive to set specified SDM to "Do Not Calibrate".

        :param message: message containing key-value pairs
        :return:
        """
        wf_request = self.info.lookup_workflow_request(request_id)

        if wf_request.workflow_name == ArchiveWorkflows.DO_NOT_CALIBRATE.value:
            logger.info(f"SENDING 'DO NOT CALIBRATE' MESSAGE to AAT for request #{request_id}!")
            dnc_msg = ArchiveMessageArchitect(
                routing_key="qa-script.fail", request=request_id, sdm_id=sdm_id
            ).compose_message("qa_fail")
            self.archive_messenger.send_message(**dnc_msg)

    def execute(self, request: WorkflowRequest):
        """
        Execute a workflow per the supplied parameters.

        :param request: The Workflow Request to execute
        :param files: additional files to execute with request
        :return:
        """
        # look up workflow, returns workflow
        logger.info("Looking up workflow %s", request.workflow_name)
        definition = self.info.lookup_workflow_definition(request.workflow_name)

        # create a temporary directory if processing directory is not supplied,
        # needs to exist before template rendering
        temp_folder = self._make_temp_directory(request) if not request.results_dir else Path(request.results_dir)

        # Is this a remote execution workflow?
        injector = CapoInjector(request)
        remote = injector.is_remote_workflow()
        run_cv = injector.is_naasc_workflow()

        # if remote is true, create a capo subspace file in the request's directory
        if remote:
            logger.info("Creating Subspace File...")
            injector.subspace_to_directory()
            request.argument["profile"] = injector.profile

        # render templates to files, returns list of rendered files
        logger.info("Rendering files")

        # Aggrandize the argument with the workflow request ID and results directory
        request.argument["request_id"] = request.workflow_request_id
        request.argument["spool_dir"] = request.results_dir

        # render all the templates
        if request.argument["need_project_metadata"] is True:
            self.determine_multiple_productfetch(request)
            templated_files = self._render_with_metadata(request, temp_folder, definition)
            # catch aat_wrest failure and abort workflow
            if isinstance(templated_files, WorkflowRequest):
                return request
        else:
            self.determine_multiple_productfetch(request)
            if remote:
                request.argument["ramInGb"] = self.processing_settings.ramInGb

            templated_files = definition.render_templates(request.argument)

        # Combine the templated with the request files.
        workflow_files = self._determine_usable_files(request, templated_files)

        # prepare files for condor execution
        self._prepare_files_for_condor(workflow_files, temp_folder)
        self.info.save_request(request)

        # execute condor and retrieve log file
        log_file = self._execute_prepared(Path(request.results_dir), run_cv)

        # start reading the log file (background it)
        logger.info("Running wf_monitor.")
        subprocess.Popen(
            ["wf_monitor", log_file, str(request.workflow_request_id)],
            cwd=request.results_dir,
        )
        logger.info("Monitor running. Moving along...")

        return request

    def _make_temp_directory(self, request: WorkflowRequest) -> Path:
        """
        Create tmp folder for workflow

        :return: path to tmp folder
        """
        vlapipe_id = 6000
        temp_folder_path = self.processing_settings.rootDirectory
        if not os.path.isdir(temp_folder_path):
            os.mkdir(temp_folder_path)
        temp_folder = Path(mkdtemp(dir=temp_folder_path))
        os.chown(temp_folder, vlapipe_id, vlapipe_id)
        logger.info("Settled on temp folder %s", temp_folder)
        # set new directory as result_dir
        request.results_dir = temp_folder.__str__()
        return temp_folder

    def determine_multiple_productfetch(self, request: WorkflowRequest):
        """
        Are we fetching more than one product?

        :param request: the workflow request
        :return:
        """
        if "need_data" in request.argument:
            if request.argument["need_data"] is True:
                spl_list = request.argument["product_locator"].split(",")
                fetcher_string = " --product-locator"
                request.argument["product_locator"] = fetcher_string + fetcher_string.join(
                    " " + spl for spl in spl_list
                )
                request.argument["ramInGb"] = self.processing_settings.ramInGb

    def _render_with_metadata(self, wf_request: WorkflowRequest, tempdir: Path, wf_definition: Workflow):
        name = wf_request.workflow_name
        if "calibration" in name:
            wrest_type = "-sc"
            argument = wf_request.argument["product_locator"]
            argument2 = []
        elif "restore" in name:
            wrest_type = "-r"
            argument = wf_request.argument["product_locator"]
            argument2 = [wf_request.argument["cal_locator"]]
        elif "imaging" in name:
            wrest_type = "-si"
            argument = wf_request.argument["sdmId"]
            argument2 = []
        elif "carta" in name:
            wrest_type = "--carta"
            argument = wf_request.argument["product_locator"].split()[1]
            argument2 = []
        elif "ingest" in name:
            wrest_type = "-aux"
            parent_req = self.info.lookup_workflow_request(int(wf_request.argument["parent_wf_request_id"]))
            eb = (
                parent_req.argument["product_locator"]
                if "product_locator" in parent_req.argument
                else parent_req.argument["sdmId"]
            )
            argument = eb
            argument2 = []
        else:
            logger.info(f"No wrester found for workflow {name}. Does it actually require metadata?")
            return wf_request

        logger.info(f" workflow {name} has wrest option: {wrest_type}")
        with subprocess.Popen(
            ["aat_wrest", wrest_type, argument] + argument2,
            stdout=subprocess.PIPE,
        ) as wf_metadata:
            wf_json = wf_metadata.communicate()[0]
            logger.info(wf_json.decode())
            if "ERROR" not in wf_json.decode():
                merged_args = {
                    **json.loads(wf_json),
                    **wf_request.argument,
                    "relative_path": str(tempdir).split("/")[-1],
                    "root_directory": self.processing_settings.rootDirectory,
                    "ramInGb": self.processing_settings.ramInGb,
                }
                return wf_definition.render_templates(merged_args)
            else:
                logger.error(wf_json.decode())
                logger.info("SENDING WORKFLOW FAIL MESSAGE!")
                failed_msg = WorkflowMessageArchitect(request=wf_request).compose_message("workflow_failed")
                self.messenger.send_message(**failed_msg)
                return wf_request

    def _determine_usable_files(self, request: WorkflowRequest, templated_files: List[WorkflowRequestFile]):
        # Override templates if user supplied file has same name and is a valid input file
        usable_templates = []
        usable_files = []
        forbidden_templates = self._get_forbidden_templates_list(request.workflow_name)
        rq_names = []

        # create list of filenames of provided request files for easy sorting
        for file in request.files:
            rq_names.append(file.filename)

        for template in templated_files:
            # check if there is not a newer version supplied, add the template
            if template.filename not in rq_names:
                usable_templates.append(template)

            # check if a newer version was supplied and is not in the list of restricted
            # access templates
            if template.filename in rq_names and template.filename in forbidden_templates:
                usable_templates.append(template)
                logger.info(
                    f"Cannot append user file {request.files[template.filename]} to "
                    f"workflow request #{request.workflow_request_id}. "
                    f"Users are not allowed to override required system templates."
                )

        # check that supplied files are not in conflict with the sorted templates
        for file in request.files:
            if file.filename not in usable_templates:
                usable_files.append(file)

        for file in usable_templates:
            self.info.save_file(request=request, filename=file.filename, content=file.content)

        return usable_templates + usable_files

    @staticmethod
    def _prepare_files_for_condor(files: List[WorkflowRequestFile], temp_folder: Path):
        """
        Place the files for Condor into a new temp directory and returns the directory.

        :param files:  a dictionary of filename -> content
        :return:       a Path
        """
        # 1. spool each of the temp files to tmp directory
        for file in files:
            path = temp_folder / file.filename
            logger.info("Writing %s", path)
            path.write_bytes(file.content)

        # 2. make any scripts in there executable
        for file in temp_folder.glob("*.sh"):
            logger.info("Making %s executable", file)
            file.chmod(file.stat().st_mode | stat.S_IEXEC)

    def _execute_prepared(self, folder: Path, run_cv: bool) -> Path:
        """
        Execute HTCondor using the named folder as the source of the files.

        :param folder:  the path to the folder to execute
        :return:        the path to the log file
        """
        logger.info("executing on folder %s", folder)

        # some file in here should end in .dag; that file is our dagman input
        dag_files = list(folder.glob("*.dag"))
        if dag_files:
            logger.info("dagman file %s exists.", dag_files[0])
            return self._submit_dag(folder, dag_files[0], run_cv)
        else:
            job_file = list(folder.glob("*.condor"))[0]
            logger.info("condor file %s exists.", job_file)
            return self._submit_job(folder, job_file, run_cv)

    def _submit_job(self, folder: Path, job_file: Path, run_cv: bool) -> Path:
        """
        Submit job file to HTCondor

        :param folder: Folder to execute workflow in
        :param job_file: Path to job submit file
        :return: PAth to workflow log file
        """
        # ensure the log file exists
        logfile = self._get_job_logfile_name(job_file)
        logger.info("log file %s exists.", logfile)
        logfile.touch()

        # submit
        logger.info("submitting job to condor...")
        subprocess.run(
            ["condor_submit", str(job_file)],
            cwd=str(folder.absolute()),
            preexec_fn=self._switch_to_submituser(run_cv),
        )

        # return the logfile
        return logfile

    def _submit_dag(self, folder: Path, dag_file: Path, run_cv: bool) -> Path:
        """
        Submit DAG file to HTCondor

        :param folder: Folder to execute workflow in
        :param dag_file: Path to DAG submit file
        :return: Path to workflow log file
        """
        # ensure the log file exists
        logfile = self._get_dag_logfile_name(dag_file)
        logger.info("log file %s exists.", logfile)
        logfile.touch()

        # submit
        logger.info("submitting DAG to condor...")
        subprocess.run(
            ["condor_submit_dag", "-f", str(dag_file)],
            cwd=str(folder.absolute()),
            preexec_fn=self._switch_to_submituser(run_cv),
        )

        # return the logfile
        return logfile

    @staticmethod
    def _switch_to_submituser(run_cv: bool):
        """
        Helper function that is meant to be called as a `preexec_fn` for `subprocess.run`;
        sets gid and uid of subprocess to those of the submituser so that jobs can be
        properly submitted
        """
        submituser_gid = 6000
        submituser_uid = 6000

        os.setgid(submituser_gid)
        os.setuid(submituser_uid)

    @staticmethod
    def _get_job_logfile_name(jobfile_name: Path) -> Path:
        """
        Read HTCondor job file and get the log file name, if it exists

        :param jobfile_name: Job file name
        :return: Log file name or "{job-name}.log"
        """
        with open(jobfile_name, "rt") as jobfile:
            for line in jobfile:
                if line.startswith("log = "):
                    _, logfile_name = line.split(" = ")
                    return Path(logfile_name.strip())
        return Path(f"{jobfile_name.stem}.log")

    @staticmethod
    def _get_dag_logfile_name(dag_file: Path) -> Path:
        """
        Return path to DAG log file, which will always be {dag-file-name}.dagman.log

        :param dag_file: Path to workflow DAG file
        :return: Path to workflow DAG log
        """
        return Path(str(dag_file) + ".dagman.log")

    @staticmethod
    def _get_forbidden_templates_list(workflow_name: str):
        return [
            workflow_name + ".sh",
            workflow_name + ".condor",
            workflow_name + ".dag",
        ]

    def announce_qa(self, workflow_request_id: int, msg_type: str):
        """
        Announce this QA ready/pass/fail event
        :param workflow_request_id: the workflow request ID
        :param msg_type: ready, pass, or fail
        :return:
        """
        logger.info(f"ANNOUNCING QA {msg_type.upper()} for request #{workflow_request_id}!")
        wf_request = self.info.lookup_workflow_request(workflow_request_id)

        qa_event_msg = WorkflowMessageArchitect(request=wf_request).compose_message(msg_type)
        self.archive_messenger.send_message(**qa_event_msg)

        return qa_event_msg

    def send_forced_fail(self, wf_request_id: int):
        """
        manually sending a fail message

        :param wf_request_id: id of request to fail
        :return:
        """
        wf_request = self.info.lookup_workflow_request(wf_request_id)
        fail_msg = WorkflowMessageArchitect(request=wf_request).compose_message("workflow_failed")
        self.messenger.send_message(**fail_msg)


class WorkflowMessageHandler:
    """Handles workflow messages. No, really!"""

    def __init__(self, info: WorkflowInfoIF):
        self.info = info
        self.message_router = Router("workflow")
        self.message_router.register(self)
        self.messenger = MessageSender("workflow")
        self.archive_messenger = MessageSender("archive")

    @on_message(service="workflow", type="workflow-complete")
    def propagate_delivery(self, **message: Dict):
        """
        Deliver this message.

        :param message: message to be delivered
        :return:
        """
        logger.info("RECEIVED WORKFLOW COMPLETE MESSAGE. SENDING DELIVERY MESSAGE.")
        # look up the workflow request to get the path
        wf_request = self.info.lookup_workflow_request(message["subject"]["workflow_request_id"])

        # find the path to the delivery.json file
        delivery_file = pathlib.Path(wf_request.results_dir) / "delivery.json"

        # does it exist? if so, we need to send out a new event with its contents
        if delivery_file.exists():
            logger.info("Found delivery.json file, sending delivery notification")
            with open(delivery_file, "r") as delivery_fo:
                delivery_info = json.load(delivery_fo)

                # construct the message
                message = WorkflowMessageArchitect(
                    previous_info=message["subject"], delivery_info=delivery_info
                ).compose_message("delivery")

                # send the message
                self.messenger.send_message(**message)
        else:
            logger.info("No delivery.json found")

    @on_message(service="workflow", type="workflow-complete")
    def send_ingestion_complete_message(self, **message: Dict):
        """
        Send AMQP message to notify the system that ingestion has succeeded

        :param message: workflow-complete message body
        """
        subject = message["subject"]
        wf_req_id = subject["workflow_request_id"]
        wf_request = self.info.lookup_workflow_request(wf_req_id)

        if "ingest" in wf_request.workflow_name:
            logger.info("SENDING INGESTION COMPLETE MESSAGE!")
            subject["execution_wf_id"] = wf_request.argument["parent_wf_request_id"]

            ingestion_complete_msg = WorkflowMessageArchitect(previous_info=subject).compose_message(
                "ingestion_complete"
            )
            self.messenger.send_message(**ingestion_complete_msg)

    @on_message(service="workflow")
    def on_workflow_event(self, **message: Dict):
        """
        Handle the workflow event indicated by this message.

        :param message:
        :return:
        """
        logger.info(f"RECEIVED WORKFLOW MESSAGE: {message}")
        msg_subject = message["subject"]
        if msg_subject["type"] == "WorkflowRequest":
            request_id = msg_subject["workflow_request_id"]
            request = self.info.lookup_workflow_request(request_id)

            # 2. update request record with new status
            if message["type"] == "workflow-executing":
                status = WorkflowRequestState.Running.name

                if htcondor_job_id := int(message["condor_metadata"]["condor_job_id"]):
                    # Workflow has corresponding condor job ID
                    logger.info(f"Workflow request has an HTCondor job ID of {htcondor_job_id}. Setting DB column!")
                    request.htcondor_job_id = htcondor_job_id
            elif message["type"] == "workflow-complete":
                status = WorkflowRequestState.Complete.name

                if ArchiveWorkflows.is_archive_wf(request.workflow_name):
                    self.send_archive_complete_event(**message)

                self.clean_remote_workflow(request)
                self.clean_workflow(request)
            elif message["type"] == "workflow-failed":
                status = WorkflowRequestState.Failed.name

                if ArchiveWorkflows.is_archive_wf(request.workflow_name):
                    self.send_archive_failed_event(**message)

                self.clean_remote_workflow(request)
                self.clean_workflow(request)
            elif message["type"] == "delivery" or message["type"] == "ingestion-complete":
                status = WorkflowRequestState.Complete.name

            else:
                status = "Unknown"

            logger.info(
                "Updating state on workflow request %s to %s...",
                request.workflow_request_id,
                status,
            )
            # update request record with new status
            try:
                request.update_status(status)
                transaction.commit()
            except Exception as exc:
                transaction.abort()
                logger.error(
                    f"Failed to update status on workflow request {request.workflow_request_id} to {status}: {exc}"
                )
        else:
            logger.warning(f"Message {message} does not concern a workflow request. Ignoring.")

    def send_archive_failed_event(self, **message: Dict):
        subject = message["subject"]
        wf_id = subject["workflow_request_id"]
        wf_request = self.info.lookup_workflow_request(wf_id)

        if wf_request.workflow_name == ArchiveWorkflows.CARTA.value and wf_request.argument["notify_ready"] is False:
            logger.info(f"SENDING FAILED CARTA MESSAGE to AAT Request Handler for request #{wf_id}!")
            routing_key = f"ws-workflow.carta-instance-ready.{wf_id}"
            carta_url_msg = ArchiveMessageArchitect(routing_key=routing_key, request=wf_request).compose_message(
                "carta_failed"
            )
            self.archive_messenger.send_message(**carta_url_msg)
        if wf_request.workflow_name == ArchiveWorkflows.SECI.value:
            logger.info(f"SENDING FAILED SECI MESSAGE to VLASS Manager for request #{wf_id}!")
            routing_key = f"ws-workflow.seci.{wf_id}"
            seci_msg = ArchiveMessageArchitect(routing_key=routing_key, request=wf_request).compose_message(
                "seci_failed"
            )
            self.archive_messenger.send_message(**seci_msg)

    def send_archive_complete_event(self, **message: Dict):
        subject = message["subject"]
        wf_id = subject["workflow_request_id"]
        wf_request = self.info.lookup_workflow_request(wf_id)

        if wf_request.workflow_name == ArchiveWorkflows.SECI.value:
            logger.info(f"SENDING SECI COMPLETE MESSAGE to VLASS Manager for request #{wf_id}!")
            routing_key = f"ws-workflow.seci.{wf_id}"
            seci_msg = ArchiveMessageArchitect(routing_key=routing_key, request=wf_request).compose_message(
                "seci_complete"
            )
            self.archive_messenger.send_message(**seci_msg)

    def clean_remote_workflow(self, request: WorkflowRequest):
        injector = CapoInjector(request)

        if injector.is_remote_workflow():
            logger.info("Cleaning remote workflow....")
            injector.clear_subspace()

    def clean_workflow(self, request: WorkflowRequest):
        if request.htcondor_job_id is not None:
            # Request has a set condor_job_id; unset it
            logger.info("Workflow finished. Unsetting HTCondor ID.")
            request.htcondor_job_id = None
